'''This module is about converting chunks of (X)HTML code into ODF code that
   pod can then integrate into an pod template, via a statement like
   
                      "do ... from xhtml(...)"
'''
