# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>This para should get style Signature.</p>
<ul>
 <li><p>This one shouldn't</p></li>
</ul>
<table>
 <tr><td><p>This one shouldn't, too.</p></td></tr>
</table>
<div><div class="Signature">Match</div></div>
'''
