var1 = 'VAR1 not overridden'
var2 = 'VAR2 not overridden'

def getVars():
    '''Returns a tuple to be unpacked as as many variables as there are tuple
       elements.'''
    return ('aa', 45, 'bb')
