# -*- coding: utf-8 -*-
xhtmlInput = '''
<p style="text-indent: 26px">Considérant qu’il y a lieu de fixer le traitement individuel de <strong><span style="background-color:Yellow">...</span>, né(e) le <span style="background-color:Yellow">...</span></strong> au grade <span style="background-color:Yellow">...</span>, niveau <span style="background-color:Yellow">...</span>, temps <span style="background-color:Yellow">...</span>, dont la carrière administrative est décrite à la fiche individuelle ci-après&nbsp;;</p>

<p>Vu le code de la démocratie locale et de la décentralisation&nbsp;;&nbsp;&nbsp;</p>

<table border="1" cellpadding="1" cellspacing="1" style="height:78px; width:17cm">
    <tbody>
        <tr>
            <td style="text-align:center; vertical-align:middle"><strong>FICHE INDIVIDUELLE DE L’AGENT</strong></td>
        </tr>
        <tr>
            <td style="text-align:center; vertical-align:middle">Né(e) le&nbsp;: <span style="background-color:Yellow">...</span>, Etat civil&nbsp;: <span style="background-color:Yellow">...</span></td>
        </tr>
        <tr>
            <td style="text-align:center; vertical-align:middle"><strong>SERVICES ADMISSIBLES</strong></td>
        </tr>
    </tbody>
</table>

<table border="1" cellpadding="1" cellspacing="1" style="width:17cm">
    <tbody>
        <tr>
            <td colspan="4" rowspan="1" style="background-color:rgb(255, 255, 255)">Supplément pour ancienneté garantie</td>
            <td colspan="5" rowspan="1" style="background-color:rgb(255, 255, 255)">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><u>Nature</u></td>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:bottom"><u>Période (bottom)</u></td>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>DUREE PAR<br />
            GROUPE</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>Coef-<br />
            ficient</strong></td>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>VALORISATION<br />
            BAREMIQUE</strong></td>
        </tr>
        <tr>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">Fonction</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">Employeur</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">du</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">au</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>ANS</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>MOIS</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">&nbsp;</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>ANS</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>MOIS</strong></td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td><strong>TOTAL :</strong></td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" rowspan="1" style="background-color:rgb(204, 204, 204); text-align:right; vertical-align:middle">Ancienneté pécuniaire au <span style="background-color:Yellow">...</span></td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" rowspan="1" style="background-color:rgb(204, 204, 204); text-align:right; vertical-align:middle">Augmentation les <span style="background-color:Yellow">...</span></td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
    </tbody>
</table>

<table cellspacing="1" style="width:720px">
    <tbody>
        <tr>
            <td colspan="9" style="height:23px; width:720px">FICHE INDIVIDUELLE DE L’AGENT</td>
        </tr>
        <tr>
            <td colspan="9" style="height:23px; width:720px">Né(e) le&nbsp;: ..., Etat civil&nbsp;: ...</td>
        </tr>
        <tr>
            <td colspan="9" style="height:23px; width:720px">SERVICES ADMISSIBLES</td>
        </tr>
        <tr>
            <td colspan="4" style="height:23px; width:320px">Supplément pour ancienneté garantie</td>
            <td colspan="5" style="width:400px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="2" style="height:52px; width:160px">Nature</td>
            <td colspan="2" style="width:160px">Période</td>
            <td colspan="2" style="width:160px">DUREE PAR<br />
            GROUPE</td>
            <td style="width:80px">Coef-<br />
            ficient</td>
            <td colspan="2" style="width:160px">VALORISATION<br />
            BAREMIQUE</td>
        </tr>
        <tr>
            <td style="height:23px; width:80px">Fonction</td>
            <td style="width:80px">Employeur</td>
            <td style="width:80px">du</td>
            <td style="width:80px">au</td>
            <td style="width:80px">ANS</td>
            <td style="width:80px">MOIS</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp; ANS</td>
            <td style="width:80px">MOIS</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" style="height:23px; width:560px">&nbsp;</td>
            <td style="width:80px">TOTAL :</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" style="height:23px; width:560px">Ancienneté pécuniaire au ...</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" style="height:23px; width:560px">Augmentation les ...</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
    </tbody>
</table>

<p><br />
Échelles barémiques&nbsp;: <span style="background-color:Yellow">...</span> – <span style="background-color:Yellow">...</span> – <span style="background-color:Yellow">...</span> (niveau - grade - fonction)</p>

<ul>
    <li>Minimum&nbsp;: <span style="background-color:Yellow">...</span></li>
    <li>Maximum&nbsp;: <span style="background-color:Yellow">...</span></li>
    <li>Développement&nbsp;:&nbsp;
    <ul>
        <li><span style="background-color:Yellow">.. x .. x ..</span></li>
        <li>&nbsp;</li>
    </ul>
    </li>
</ul>
'''
