/* Copyright (c) 2003-2013, CKSource - Frederico Knabben. All rights reserved.
   For licensing, see LICENSE.html or http://ckeditor.com/license */

/* From ckeditor default contents.css */
body { font-family: sans-serif, Arial, Verdana, "Trebuchet MS"; font-size: 11px;
  color: #333; background-color: #fff; margin: 10px }
.cke_editable { font-size: 12px; line-height: 1.4 }
blockquote { font-style: italic;
  font-family: Georgia, Times, "Times New Roman", serif; padding: 2px 0;
  border-style: solid; border-color: #ccc; border-width: 0 }
.cke_contents_ltr blockquote {
  padding-left: 5px; padding-right: 5px; border-left-width: 4px }
.cke_contents_rtl blockquote {
  padding-left: 5px; padding-right: 5px; border-right-width: 4px }
a { color: #0782C1 }
ol,ul,dl { *margin-right: 0px; padding: 0 30px }
h1,h2,h3,h4,h5,h6 { font-weight: normal; line-height: 1.2 }
hr { border: 0px; border-top: 1px solid #ccc }
img.right { border: 1px solid #ccc; float: right; margin-left: 15px;
            padding: 5px }
img.left { border: 1px solid #ccc; float: left; margin-right: 15px;
            padding: 5px }
pre { white-space: pre-wrap; word-wrap: break-word; -moz-tab-size: 4;
      -o-tab-size: 4; -webkit-tab-size: 4; tab-size: 4 }
.marker { background-color: Yellow }
span[lang] { font-style: italic }
figure { text-align: center; border: solid 1px #ccc; border-radius: 2px;
  background: rgba(0,0,0,0.05); padding: 10px; margin: 10px 20px;
  display: inline-block }
figure > figcaption { text-align: center; display: block }
a > img { padding: 1px; margin: 1px; border: none; outline: 1px solid #0782C1 }

/* Added by Appy */
p { margin: 0; padding: 0 0 3px 0 }
table { border-collapse: collapse; border-spacing: 0; font-size: 95% }
td, th { border: 1px solid grey; padding: 3px }
th { background-color: #f0f0f0 }
